//**************************** sample.c ******************************
//* This is a sample application program that shows you how to use the
//* API function calls of the loadable Linux device driver 'bttv' and 
//* 's609drv' for Sensoray Model 609 to develop an image-capturing 
//* application.
//*
//* Copyright (c) 2003.  Sensoray Corp.
//* Author:              Charlie X. Liu
//*
//*--------------------------------------------------------------------------
//* All rights reserved.
//*
//* Redistribution and use in source and binary forms, with or without
//* modification, are permitted provided that the following conditions
//* are met:
//*
//* 1. Redistributions of source code must retain the above copyright
//*    notice, this list of conditions and the following disclaimer.
//* 2. Redistributions in binary form must reproduce the above copyright
//*    notice, this list of conditions and the following disclaimer in the
//*    documentation and/or other materials provided with the distribution.
//* 3. All advertising materials mentioning features or use of this software
//*    must display the following acknowledgement:
//*      This product includes software developed by Charlie X. Liu
//* 4. The name of the author may not be used to endorse or promote products
//*    derived from this software without specific prior written permission.
//*
//* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
//* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//* DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
//* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
//* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
//* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
//* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
//* ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
//* POSSIBILITY OF SUCH DAMAGE.
//*--------------------------------------------------------------------------
//* This application captures 24-bit RGB images from 8 video input channels
//* and save them into 8 image files respectively.

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/types.h>

#include <linux/types.h>
#include <linux/videodev.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "s609drv.h"

#define DISPLAY_INFO		//turn on message display

int main(int argc, char **argv)
{
  int WIDTH=640, HEIGHT=480;
  //int WIDTH=320, HEIGHT=240;
  //int WIDTH=160, HEIGHT=120;
  int    n, fd, fd_609;
  char   *map, *p;
  char   tmp;
  struct video_capability vd;
  struct video_channel vc[10];
  struct video_picture vp;
  struct video_mbuf vm;
  struct video_mmap vmm[2];
  int    ch;
  WORD   vin, vout;
  int    counts;

  int outfile;
  int width, height;
  unsigned char *buffer;
  unsigned char *line_buffer;
  unsigned char header[30];
  char *device_name1, *device_name2;
  char *filename1, *filename2, *filename3, *filename4;
  char *filename5, *filename6, *filename7, *filename8;
  int h, w;

  unsigned long data;
  struct ioparam request;

  filename1 = "image1.ppm";
  filename2 = "image2.ppm";
  filename3 = "image3.ppm";
  filename4 = "image4.ppm";
  filename5 = "image5.ppm";
  filename6 = "image6.ppm";
  filename7 = "image7.ppm";
  filename8 = "image8.ppm";


  if(argc>2) { WIDTH=atoi(argv[1]); HEIGHT=atoi(argv[2]); }

  /* Video4Linux open devices */
  device_name1 = "/dev/video0";
  device_name2 = "/dev/s609";

  //open bttv
  if((fd = open(device_name1, O_RDWR)) < 0) {
    perror("bttv open");
    return -1;
  }
  //open s609drv
  if((fd_609 = open(device_name2, O_RDWR)) < 0) {
    perror("s609drv open");
    return -1;
  }

  /* Video4Linux get device info */
  if(ioctl(fd, VIDIOCGCAP, &vd)<0) {
    perror("ioctl(VIDIOCGCAP)");
    return -1;
  }
  #ifdef DISPLAY_INFO
  fprintf(stderr, "vd.name: \"%s\"\n", vd.name);
  fprintf(stderr, "vd.type      = 0x%08x", vd.type);
  if(vd.type&VID_TYPE_CAPTURE)    fprintf(stderr, " CAPTURE");
  if(vd.type&VID_TYPE_TUNER)      fprintf(stderr, " TUNER");
  if(vd.type&VID_TYPE_TELETEXT)   fprintf(stderr, " TELETEXT");
  if(vd.type&VID_TYPE_OVERLAY)    fprintf(stderr, " OVERLAY");
  if(vd.type&VID_TYPE_CHROMAKEY)  fprintf(stderr, " CHROMAKEY");
  if(vd.type&VID_TYPE_CLIPPING)   fprintf(stderr, " CLIPPING");
  if(vd.type&VID_TYPE_FRAMERAM)   fprintf(stderr, " FRAMERAM");
  if(vd.type&VID_TYPE_SCALES)     fprintf(stderr, " SCALES");
  if(vd.type&VID_TYPE_MONOCHROME) fprintf(stderr, " MONOCHROME");
  if(vd.type&VID_TYPE_SUBCAPTURE) fprintf(stderr, " SUBCAPTURE");
  fprintf(stderr, "\n");
  fprintf(stderr, "vd.channels  = %d\n", vd.channels);
  fprintf(stderr, "vd.audios    = %d\n", vd.audios);
  fprintf(stderr, "vd.maxwidth  = %d\n", vd.maxwidth);
  fprintf(stderr, "vd.maxheight = %d\n", vd.maxheight);
  fprintf(stderr, "vd.minwidth  = %d\n", vd.minwidth);
  fprintf(stderr, "vd.minheight = %d\n", vd.minheight);
  fprintf(stderr, "\n");
  #endif

  /* adjust capture size */
  if(WIDTH>vd.maxwidth) WIDTH=vd.maxwidth;
  if(WIDTH<vd.minwidth) WIDTH=vd.minwidth;
  if(HEIGHT>vd.maxheight) HEIGHT=vd.maxheight;
  if(HEIGHT<vd.minheight) HEIGHT=vd.minheight;

  /* get internal channel info */
  for(n = 0 ; n < vd.channels ; n++) {
    vc[n].channel = n;
    if(ioctl(fd, VIDIOCGCHAN, &vc[n])<0) {
      perror("ioctl(VIDIOCGCHAN)");
      return -1;
    }
    #ifdef DISPLAY_INFO
    fprintf(stderr, "vc[%d].channel = %d\n", n, vc[n].channel);
    fprintf(stderr, "vc[%d].name    = \"%s\"\n", n, vc[n].name);
    fprintf(stderr, "vc[%d].tuners  = %d\n", n, vc[n].tuners);
    fprintf(stderr, "vc[%d].flags   = 0x%08x", n, vc[n].flags);
    if(vc[n].flags&VIDEO_VC_TUNER) fprintf(stderr, " TUNER");
    if(vc[n].flags&VIDEO_VC_AUDIO) fprintf(stderr, " AUDIO");
    fprintf(stderr, "\n");
    fprintf(stderr, "vc[%d].type    = 0x%08x", n, vc[n].type);
    if(vc[n].type&VIDEO_TYPE_TV) fprintf(stderr, " TV");
    if(vc[n].type&VIDEO_TYPE_CAMERA) fprintf(stderr, " CAMERA");
    fprintf(stderr, "\n");
    fprintf(stderr, "vc[%d].norm    = %d\n", n, vc[n].norm);
    fprintf(stderr, "\n");
    #endif
  }


  //* 0 = Composite0
  //* 1 = Composite1
  //* 2 = Composite2
  //* 3 = Composite3
  //* 4 = S-Video	not for S609
  ch = 0;	/* choose "Composite0 = 0" */
  vc[ch].norm = VIDEO_MODE_NTSC; 	/* 0:PAL 1:NTSC 2:SECAM 3:AUTO*/
  if(ioctl(fd, VIDIOCSCHAN, &vc[ch])<0) {
    perror("ioctl(VIDIOCSCHAN)");	/* switch capture to the channel */
    return -1;
  }

  /* set template (range: 0-65535) */
  vp.brightness = 32767;
  vp.hue        = 32767;
  vp.colour     = 32767;
  vp.contrast   = 32767;
  vp.whiteness  = 32767;
  vp.depth      = 24; /* color depth */
  vp.palette    = VIDEO_PALETTE_RGB24; /* mode */
  if(ioctl(fd, VIDIOCSPICT, &vp)) {
    perror("ioctl(VIDIOCSPICT)");
    return -1;
  }

  /* get mmap info */
  if(ioctl(fd, VIDIOCGMBUF, &vm)<0) {
    perror("ioctl(VIDIOCGMBUF)");
    return -1;
  }
  #ifdef DISPLAY_INFO
  fprintf(stderr, "vm.size=0x%08x\n", vm.size);
  fprintf(stderr, "vm.frames=0x%08x\n", vm.frames);
  for( n = 0 ; n < vm.frames ; n++)
    fprintf(stderr, "vm.offsets[%d]=0x%08x\n", n, vm.offsets[n]);
  fprintf(stderr, "\n");
  #endif


  /* mmap */
  if((map=mmap(0, vm.size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0))==(char *)-1){
    perror("mmap");
    return -1;
  }

  /* frame #0 capturing parameter setting */
  vmm[0].frame  = 0;
  vmm[0].width  = WIDTH;
  vmm[0].height = HEIGHT;
  vmm[0].format = VIDEO_PALETTE_RGB24;

  /* frame #1 capturing parameter setting */
  vmm[1].frame  = 1;
  vmm[1].width  = WIDTH;
  vmm[1].height = HEIGHT;
  vmm[1].format = VIDEO_PALETTE_RGB24;

  //ready to go
#if 0
    // for using push-button to start capturing
    printf("\nPush button to Start capturing...\n");
    while(1)
    {
	if(ioctl(fd, DioRD, &request)<0) {
	   perror("ioctl(DioRD)");
	   return -1;
	}
	data = request.param1;

	if (!(data<<31)) {printf("\nStart capturing...\n"); break;}
    }
#endif

  // capturing frames from video input 1-8
  printf("start capturing:\n");
  for (vin=1; vin<=8; vin++) {
    vout = 1;	// video input channel vin# --> composite 0
    request.param1 = vin;	// take v-in #
    request.param2 = vout;	// take v-out #
    if (ioctl(fd_609, XptSW, &request)<0) {
       perror("ioctl(DioWR)");
       return -1;
    }
    usleep(11000); //sleep some time for switching to be stable
                   //   11 ms is the shortest time for my machine

    /* Create the output files */
    switch (vin)
    {
      case 1:
      {
	if ((outfile = open(filename1, O_WRONLY | O_CREAT, 0666)) < 0) {
		printf("ppm file 1 open failed: \n");
		exit(1);
	}
	break;
      }
      case 2:
      {
	if ((outfile = open(filename2, O_WRONLY | O_CREAT, 0666)) < 0) {
		printf("ppm file 2 open failed: \n");
		exit(1);
	}
	break;
      }
      case 3:
      {
	if ((outfile = open(filename3, O_WRONLY | O_CREAT, 0666)) < 0) {
		printf("ppm file 3 open failed: \n");
		exit(1);
	}
	break;
      }
      case 4:
      {
	if ((outfile = open(filename4, O_WRONLY | O_CREAT, 0666)) < 0) {
		printf("ppm file 4 open failed: \n");
		exit(1);
	}
	break;
      }
      case 5:
      {
	if ((outfile = open(filename5, O_WRONLY | O_CREAT, 0666)) < 0) {
		printf("ppm file 5 open failed: \n");
		exit(1);
	}
	break;
      }
      case 6:
      {
	if ((outfile = open(filename6, O_WRONLY | O_CREAT, 0666)) < 0) {
		printf("ppm file 6 open failed: \n");
		exit(1);
	}
	break;
      }
      case 7:
      {
	if ((outfile = open(filename7, O_WRONLY | O_CREAT, 0666)) < 0) {
		printf("ppm file 7 open failed: \n");
		exit(1);
	}
	break;
      }
      case 8:
      {
	if ((outfile = open(filename8, O_WRONLY | O_CREAT, 0666)) < 0) {
		printf("ppm file 8 open failed: \n");
		exit(1);
	}
	break;
      }
      default:
        printf("invalid video input channel # !\n");
    }
    /* add in header */
    width = WIDTH;
    height = HEIGHT;
    sprintf(header, "P6\n%d %d\n255\n",width,height);
    write (outfile, header, strlen(header));
    
    /* start capturing as frame #0*/
    if(ioctl(fd, VIDIOCMCAPTURE, &(vmm[0]))<0) {
	perror("ioctl(VIDIOCMCAPTURE)");
	return -1;
    }

    /* frame #0 wait for capturing stop */
    n = 0;
    if(ioctl(fd, VIDIOCSYNC, &n)<0) {
	   perror("ioctl(VIDIOCSYNC)");
	   return -1;
    }

    /* save the RGB data of frame0 to PPM file */
    /* save this one line at a time */
    line_buffer =(unsigned char *)malloc( width *3 * sizeof(unsigned char));
    buffer = map + vm.offsets[0];
    p = buffer;
    for (h = 0; h < height; h++) {
	for (w = 0; w < width; w++) {
		line_buffer[(w*3) + 2] = *p++;
		line_buffer[(w*3) + 1] = *p++;
		line_buffer[(w*3) + 0] = *p++;
	}
	write(outfile,line_buffer, width*3);
    }

    printf ("\ncapturing from video input channel %d is done and the image has been saved into 'image%d.ppm' file.", vin, vin);

    close(outfile);
    free(line_buffer);

  }

  /* close devices */
  close(fd);
  close(fd_609);

  printf ("\n\ngo get the images. ^_^ \n\n");

  return 0;

}






